from ftw.labels.portlets.interfaces import ILabelJarPortlet
from ftw.labels.portlets.interfaces import ILabelingPortlet
from plone.app.portlets.portlets.base import Assignment
from zope.interface import implements


class LabelJarAssignment(Assignment):
    implements(ILabelJarPortlet)

    @property
    def title(self):
        return 'ftw.labels: Label Jar Portlet'


class LabelingAssignment(Assignment):
    implements(ILabelingPortlet)

    @property
    def title(self):
        return 'ftw.labels: Labeling Portlet'



import unicodedata


def make_sortable(text):
    """Converts a string to a sortable string by lowercasing
    it and removing diacritics.
    """
    if isinstance(text, str):
        text = text.decode('utf-8')
    if not isinstance(text, unicode):
        return text
    text = text.lower()
    normalized = unicodedata.normalize('NFKD', text)
    text = u''.join([c for c in normalized if not unicodedata.combining(c)])
    text = text.encode('utf-8')
    return text



from Products.CMFCore.utils import getToolByName
from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.interfaces import ILabeling
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from unittest2 import TestCase


class TestCatalogIndex(TestCase):
    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        self.catalog = getToolByName(self.portal, 'portal_catalog')

    def test_index_is_registered(self):
        self.assertIn('labels', self.catalog.indexes())

    def test_updating_index(self):
        root = create(Builder('label root')
                      .with_labels(('Question', 'blue'),
                                   ('Bugs', 'red'),
                                   ('Enhancements', 'green')))

        page = create(Builder('labelled page')
                      .within(root)
                      .with_labels('question', 'bugs'))

        self.assertFalse(self.index_data_for(page).get('labels'))

        ILabeling(page).update(['bugs', 'question'])
        page.reindexObject(idxs=['labels'])

        self.assertItemsEqual(
            ['bugs', 'question'],
            self.index_data_for(page).get('labels'))

    def index_data_for(self, obj):
        rid = self.catalog.getrid('/'.join(obj.getPhysicalPath()))
        return self.catalog.getIndexDataForRID(rid)



from ftw.labels.interfaces import ILabeling
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelSupport
from ftw.labels.utils import make_sortable
from persistent.list import PersistentList
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.interface import implements


ANNOTATION_KEY = 'ftw.labels:labeling'


class Labeling(object):
    implements(ILabeling)
    adapts(ILabelSupport)

    def __init__(self, context):
        self.context = context
        self.jar = ILabelJar(self.context)

    def update(self, label_ids):
        available_labels = self.jar.storage.keys()

        for label_id in label_ids:
            if label_id not in available_labels:
                raise LookupError(
                    'Cannot activate label: '
                    'the label "{0}" is not in the label jar. '
                    'Following labels ids are available: {1}'.format(
                        label_id, ', '.join(available_labels)))

        # Do not replace self.storage, since it is a PersistentList!
        self.storage[:] = label_ids

    def active_labels(self):
        labels = []
        for label_id in self.storage:
            try:
                labels.append(self.jar.get(label_id))
            except KeyError:
                pass
        return sorted(labels, key=lambda cls: make_sortable(cls['title']))

    def available_labels(self):
        for label in self.jar.list():
            label['active'] = (label.get('label_id') in self.storage)
            yield label

    @property
    def storage(self):
        if getattr(self, '_storage', None) is None:
            annotation = IAnnotations(self.context)
            if ANNOTATION_KEY not in annotation:
                annotation[ANNOTATION_KEY] = PersistentList()
            self._storage = annotation[ANNOTATION_KEY]
        return self._storage



from OFS.interfaces import IApplication
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.jar import LabelJar
from ftw.labels.testing import ADAPTERS_ZCML_LAYER
from ftw.testing import MockTestCase


class TestJarDiscovery(MockTestCase):
    layer = ADAPTERS_ZCML_LAYER

    def test_adapting_root_returns_jar(self):
        root = self.providing_stub(ILabelRoot)
        self.replay()
        jar = ILabelJar(root)
        self.assertIsInstance(jar, LabelJar)

    def test_walks_up_the_acquisition_for_finding_jar(self):
        root = self.providing_stub(ILabelRoot)
        folder = self.set_parent(self.stub(), root)
        document =  self.set_parent(self.stub(), folder)
        self.replay()

        jar = ILabelJar(document)
        self.assertIsInstance(jar, LabelJar)
        self.assertEquals(root, jar.context)

    def test_raise_when_app_is_reached(self):
        app = self.providing_stub(IApplication)
        document = self.set_parent(self.stub(), app)
        self.replay()

        with self.assertRaises(LookupError) as cm:
            ILabelJar(document)

        self.assertEquals(
            'Could not find ILabelJar on any parents.'
            ' No parent seems to provide ILabelRoot.',
            str(cm.exception))



from ftw.builder import builder_registry
from ftw.builder.archetypes import ArchetypesBuilder
from ftw.labels.interfaces import ILabelJarChild
from ftw.labels.interfaces import ILabeling
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.interfaces import ILabelSupport
import transaction


class LabelRootBuilder(ArchetypesBuilder):

    portal_type = 'Folder'

    def __init__(self, *args, **kwargs):
        super(LabelRootBuilder, self).__init__(*args, **kwargs)
        self.providing(ILabelRoot)
        self.labels = []

    def with_labels(self, *labels):
        self.labels.extend(labels)
        return self

    def after_create(self, obj):
        super(LabelRootBuilder, self).after_create(obj)
        jar = ILabelJar(obj)
        for title, color in self.labels:
            jar.add(title, color)

        if self.session.auto_commit:
            transaction.commit()


builder_registry.register('label root', LabelRootBuilder)


class LabelDisplayBuilder(ArchetypesBuilder):

    portal_type = 'Folder'

    def __init__(self, *args, **kwargs):
        super(LabelDisplayBuilder, self).__init__(*args, **kwargs)
        self.providing(ILabelJarChild)


builder_registry.register('label display', LabelDisplayBuilder)


class LabelledPageBuilder(ArchetypesBuilder):

    portal_type = 'Document'

    def __init__(self, *args, **kwargs):
        super(LabelledPageBuilder, self).__init__(*args, **kwargs)
        self.providing(ILabelSupport)
        self.activated_label_ids = []

    def with_labels(self, *label_ids):
        self.activated_label_ids = label_ids
        return self

    def after_create(self, obj):
        super(LabelledPageBuilder, self).after_create(obj)

        ILabeling(obj).update(self.activated_label_ids)
        if self.session.auto_commit:
            transaction.commit()


builder_registry.register('labelled page', LabelledPageBuilder)



COLORS = [
    'red',
    'orange',
    'yellow',
    'green',
    'cadetblue',
    'blue',
    'purple',
    'cornflowerblue',
    'red-light',
    'orange-light',
    'yellow-light',
    'green-light',
    'cadetblue-light',
    'cornflowerblue-light',
    'blue-light',
    'purple-light',
]



from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.interfaces import ILabelSupport
from ftw.labels.interfaces import ILabeling
from ftw.labels.testing import ADAPTERS_ZCML_LAYER
from ftw.testing import MockTestCase
from zope.annotation import IAttributeAnnotatable
from zope.component import queryAdapter


def label_titles(labels):
    return [label['title'] for label in labels]


class TestLabeling(MockTestCase):
    layer = ADAPTERS_ZCML_LAYER

    def setUp(self):
        super(TestLabeling, self).setUp()
        self.root = self.providing_stub([ILabelRoot, IAttributeAnnotatable])
        self.document = self.providing_stub([ILabelSupport,
                                             IAttributeAnnotatable])
        self.set_parent(self.document, self.root)
        self.replay()
        self.jar = ILabelJar(self.root)

    def test_adapter(self):
        self.assertTrue(
            queryAdapter(self.document, ILabeling),
            'The labeling adapter is not registered for ILabeling')

    def test_available_labels(self):
        self.jar.add('Question', '#00FF00')
        labeling = ILabeling(self.document)
        self.assertEqual(
            [{'label_id': 'question',
             'title': 'Question',
             'color': '#00FF00',
             'active': False}],
            list(labeling.available_labels()))

    def test_available_labels_empty(self):
        labeling = ILabeling(self.document)
        self.assertEqual([], list(labeling.available_labels()))

    def test_available_label(self):
        self.jar.add('Question', '#00FF00')
        labeling = ILabeling(self.document)

        labeling.update(['question'])
        self.assertEqual(
            [{'label_id': 'question',
              'title': 'Question',
              'color': '#00FF00',
              'active': True}],
            list(labeling.available_labels()))

    def test_update__enable_labels(self):
        self.jar.add('Bug', 'red')
        self.jar.add('Question', 'green')
        self.jar.add('Feature', 'purple')

        labeling = ILabeling(self.document)
        self.assertEqual([], labeling.active_labels())

        labeling.update(['bug', 'feature'])
        self.assertItemsEqual(['Bug', 'Feature'],
                              label_titles(labeling.active_labels()))

    def test_update__disable_labels(self):
        self.jar.add('Bug', 'red')
        self.jar.add('Question', 'green')
        self.jar.add('Feature', 'purple')

        labeling = ILabeling(self.document)
        labeling.update(['bug', 'question', 'feature'])
        self.assertItemsEqual(['Bug', 'Feature', 'Question'],
                              label_titles(labeling.active_labels()))

        labeling.update(['bug'])
        self.assertItemsEqual(['Bug'],
                              label_titles(labeling.active_labels()))

        labeling.update([])
        self.assertEqual([], labeling.active_labels())

    def test_update_raises_LookupError_when_label_not_in_jar(self):
        self.assertEqual(0, len(self.jar.list()))
        self.jar.add('Question', '')
        labeling = ILabeling(self.document)
        with self.assertRaises(LookupError) as cm:
            labeling.update(['something'])

        self.assertEqual(
            'Cannot activate label: the label'
            ' "something" is not in the label jar. '
            'Following labels ids are available: question',
            str(cm.exception))

    def test_active_labels(self):
        self.jar.add('Question', '')
        self.jar.add('Bug', '')
        self.jar.add('Duplicate', '')

        labeling = ILabeling(self.document)
        labeling.update(['bug'])
        self.assertEqual(
            [{'label_id': 'bug',
              'title': 'Bug',
              'color': ''}],
            labeling.active_labels())

    def test_active_labels_is_sorted(self):
        self.jar.add('Zeta-0', '')
        self.jar.add('zeta-1', '')
        self.jar.add('alpha-0', '')
        self.jar.add('\xc3\x84lpha-1', '')
        self.jar.add('Alpha-2', '')

        labeling = ILabeling(self.document)
        labeling.update([
                'zeta-0',
                'zeta-1',
                'alpha-0',
                'alpha-1',
                'alpha-2',
                ])

        self.assertEqual(
            ['alpha-0', '\xc3\x84lpha-1', 'Alpha-2', 'Zeta-0', 'zeta-1'],
            [label.get('title') for label in labeling.active_labels()])

    def test_active_labels_filters_deleted_labels(self):
        self.jar.add('Question', 'blue')
        self.jar.add('Bug', 'red')

        labeling = ILabeling(self.document)
        labeling.update(['question', 'bug'])

        self.jar.remove('bug')

        self.assertEqual(
            [{'label_id': 'question',
             'title': 'Question',
             'color': 'blue'}],
            list(labeling.active_labels()))



from asserts import assert_true
from ftw.testbrowser import browser


def portlet():
    return browser.css('.labelJarPortlet').first_or_none


def labels():
    assert_true(portlet())
    return dict((label.text, label_color(label))
                for label in portlet().css('.labelListing .labelColor .labelTitle'))


def label_color(label_li):
    return label_li.parent().css('.labelColor').first.attrib.get('data-color')



from ftw.labels.config import COLORS
from ftw.labels.interfaces import ILabelJarChild
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.portlets.assignments import LabelJarAssignment
from plone.app.portlets.portlets.base import NullAddForm
from plone.app.portlets.portlets.base import Renderer
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile


class AddForm(NullAddForm):

    def create(self):
        return LabelJarAssignment()


class Renderer(Renderer):
    render = ViewPageTemplateFile('labeljar.pt')

    @property
    def available(self):
        if 'portal_factory' in self.context.absolute_url():
            return False
        if ILabelJarChild.providedBy(self.context):
            if not self.label_root_obj:
                return False
        elif not ILabelRoot.providedBy(self.context):
            return False
        if not self.can_edit and not self.labels:
            return False
        return True

    @property
    def labels(self):
        return ILabelJar(self.label_root_obj).list()

    @property
    def colors(self):
        return COLORS

    @property
    def can_edit(self):
        mtool = getToolByName(self.context, 'portal_membership')
        return mtool.checkPermission(
            'ftw.labels: Manage Labels Jar', self.context)

    @property
    def label_root_obj(self):
        try:
            return ILabelJar(self.context)
        except LookupError:
            return None



from asserts import assert_true
from ftw.testbrowser import browser


def portlet():
    return browser.css('.labelingPortlet').first_or_none


def active_labels():
    assert_true(portlet(), 'labeling portlet is not visible')
    return portlet().css('.activeLabels .labelItem').text


def form():
    node = portlet().css('.updateLabeling').first_or_none
    assert_true(node, 'The updateLabeling form is missing.')
    return node


def form_checkboxes():
    result = {}
    for input in form().inputs:
        if not input.label:
            continue  # submit button
        result[input.label.text] = bool(input.attrib.get('checked'))
    return result



from Products.Five.browser import BrowserView
from ftw.labels.interfaces import ILabeling


class Labeling(BrowserView):

    def update(self):
        """Update activated labels.
        """
        labeling = ILabeling(self.context)
        activate_labels = self.request.form.get('activate_labels', [])
        labeling.update(activate_labels)
        self.context.reindexObject(idxs=['labels'])
        return self._redirect()

    def _redirect(self):
        response = self.request.RESPONSE
        referer = self.request.get('HTTP_REFERER')
        if referer and referer is not 'localhost':
            response.redirect(referer)
        else:
            response.redirect(self.context.absolute_url())



from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabeling
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from unittest2 import TestCase


class TestTestingBuidlers(TestCase):
    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def test_label_root_builder(self):
        root = create(Builder('label root')
                      .with_labels(('Questions', 'blue'),
                                   ('Bugs', 'red')))

        self.assertItemsEqual(
            [{'label_id': 'questions',
              'title': 'Questions',
              'color': 'blue'},
             {'label_id': 'bugs',
              'title': 'Bugs',
              'color': 'red'}],
            ILabelJar(root).list())

    def test_labelled_page_builder(self):
        root = create(Builder('label root')
                      .with_labels(('Questions', 'blue'),
                                   ('Bugs', 'red'),
                                   ('Enhancements', 'green')))

        page = create(Builder('labelled page')
                      .within(root)
                      .with_labels('questions', 'bugs'))

        self.assertItemsEqual(
            [{'label_id': 'questions',
              'title': 'Questions',
              'color': 'blue'},
             {'label_id': 'bugs',
              'title': 'Bugs',
              'color': 'red'}],
            ILabeling(page).active_labels())



from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from ftw.labels.tests.pages import jarportlet
from ftw.testbrowser import browsing
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase


class LabelJarPortletFunctionalTest(TestCase):

    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

    @browsing
    def test_portlet_is_disabled_per_default(self, browser):
        browser.visit()
        self.assertFalse(jarportlet.portlet())

    @browsing
    def test_protlet_is_enabled_if_ILabelRoot_is_provided(self, browser):
        folder = create(Builder('label root'))
        browser.login().visit(folder)
        self.assertTrue(jarportlet.portlet())

    @browsing
    def test_protlet_is_enabled_if_ILabelJarChild_and_parent_is_ILabelRoot(self, browser):
        folder = create(Builder('label root'))
        subfolder = create(Builder('label display').within(folder))

        browser.login().visit(subfolder)
        self.assertTrue(jarportlet.portlet())

    @browsing
    def test_invisible_ILabelJarChild_without_parent_ILabelRoot(self, browser):
        folder = create(Builder('label display'))

        browser.login().visit(folder)
        self.assertFalse(
            jarportlet.portlet(),
            "The folder can't show the portlet because "
            "it has no parent label root folder")

    @browsing
    def test_invisible_if_no_ILabelJarChild_but_parent_with_ILabelRoot(self, browser):
        folder = create(Builder('label root'))
        subfolder = create(Builder('folder').within(folder))

        browser.login().visit(subfolder)
        self.assertFalse(
            jarportlet.portlet(),
            "The folder shouldn't show the portlet because "
            "it has no ILabelJarChild marker interface")

    @browsing
    def test_invisible_for_unprivileged_users_when_empty(self, browser):
        folder = create(Builder('label root'))
        browser.visit(folder)
        self.assertFalse(
            jarportlet.portlet(),
            'Anonymous users should not see labels portlet when it is empty.')

    @browsing
    def test_list_all_labels_in_the_jar(self, browser):
        folder = create(Builder('label root')
                        .with_labels(('Label 1', ''), ('Label 2', '')))
        browser.visit(folder)
        self.assertItemsEqual(
            ['Label 1', 'Label 2'],
            jarportlet.labels().keys())

    @browsing
    def test_list_same_labels_in_the_labeldisplay_folder(self, browser):
        folder = create(Builder('label root')
                        .with_labels(('Label 1', ''), ('Label 2', '')))

        subfolder = create(Builder('label display').within(folder))

        browser.visit(subfolder)
        self.assertItemsEqual(
            ['Label 1', 'Label 2'],
            jarportlet.labels().keys())

    @browsing
    def test_add_color_to_each_listing_item(self, browser):
        folder = create(Builder('label root').with_labels(('James', 'red')))
        browser.visit(folder)
        self.assertEquals({'James': 'red'}, jarportlet.labels())

    @browsing
    def test_create_label(self, browser):
        folder = create(Builder('label root'))
        browser.login().visit(folder)
        self.assertEquals({}, jarportlet.labels())

        browser.forms['create-label'].fill({
                'title': 'Question',
                'color': 'purple'}).submit()

        self.assertEquals({'Question': 'purple'}, jarportlet.labels())

    @browsing
    def test_user_with_permission_can_view_jar_edit_elements(self, browser):
        folder = create(Builder('label root').with_labels(('James', 'red')))
        editor = create(Builder('user').with_roles('Editor'))
        browser.login(editor).visit(folder)

        self.assertTrue(browser.css('#create-label'))
        self.assertTrue(browser.css('.editLabelLink'))

    @browsing
    def test_user_without_permission_cant_view_jar_edit_elements(self, browser):
        folder = create(Builder('label root').with_labels(('James', 'red')))
        reader = create(Builder('user').with_roles('Reader'))
        browser.login(reader).visit(folder)

        self.assertFalse(browser.css('#create-label'))
        self.assertFalse(browser.css('.editLabelLink'))



from ftw.labels.interfaces import ILabelSupport
from ftw.labels.interfaces import ILabeling
from operator import itemgetter
from plone.indexer.decorator import indexer


@indexer(ILabelSupport)
def labels(obj):
    labeling = ILabeling(obj)
    return map(itemgetter('label_id'), labeling.active_labels())



from Products.CMFCore.utils import getToolByName
import logging


LOGGER = logging.getLogger('ftw.labels')
INDEXES = [('labels', 'KeywordIndex')]


def import_various(context):
    action = context.readDataFile('ftw.labels.setuphandlers.txt')
    if action is None:
        return

    action = action.strip()
    site = context.getSite()

    if action == 'install':
        add_catalog_indexes(site)

    elif action == 'uninstall':
        remove_catalog_indexes(site)


def add_catalog_indexes(context):
    """Method to add our wanted indexes to the portal_catalog.

    @parameters:

    When called from the import_various method below, 'context' is
    the plone site and 'logger' is the portal_setup logger.  But
    this method can also be used as upgrade step, in which case
    'context' will be portal_setup and 'logger' will be None.
    """

    catalog = getToolByName(context, 'portal_catalog')
    indexes = catalog.indexes()
    # Specify the indexes you want, with ('index_name', 'index_type')
    indexables = []
    for name, meta_type in INDEXES:
        if name not in indexes:
            catalog.addIndex(name, meta_type)
            indexables.append(name)
            LOGGER.info("Added %s for field %s.", meta_type, name)
    if len(indexables) > 0:
        LOGGER.info("Indexing new indexes %s.", ', '.join(indexables))
        catalog.manage_reindexIndex(ids=indexables)


def remove_catalog_indexes(context):
    catalog = getToolByName(context, 'portal_catalog')
    indexes = catalog.indexes()

    for name, meta_type in INDEXES:
        if name in indexes:
            catalog.delIndex(name)



from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from ftw.testbrowser import browsing
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase


class TestLabelingViewlet(TestCase):
    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        self.root = create(Builder('label root')
                           .with_labels(('Label 1', ''), ('Label 2', '')))
        self.document = create(Builder('labelled page').within(self.root))

    @browsing
    def test_no_labeling_viewlet_if_no_available_labels(self, browser):
        no_label_root = create(Builder('label root'))
        no_label_document = create(Builder('labelled page').within(no_label_root))
        browser.open(no_label_document)
        self.assertFalse(browser.css('#labeling-viewlet'))

    @browsing
    def test_show_labeling_viewlet_if_available_labels(self, browser):
        browser.open(self.document)
        self.assertTrue(browser.css('#labeling-viewlet'))

    @browsing
    def test_users_with_permission_can_view_manage_link(self, browser):
        editor = create(Builder('user').with_roles('Editor'))
        browser.login(editor).open(self.document)
        self.assertTrue(browser.css('#labeling-viewlet #toggle-label-form'))

    @browsing
    def test_users_without_permission_cant_view_manage_link(self, browser):
        reader = create(Builder('user').with_roles('Reader'))
        browser.login(reader).open(self.document)
        self.assertFalse(browser.css('#labeling-viewlet #toggle-label-form'))



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testing.layer import ComponentRegistryLayer
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from zope.configuration import xmlconfig
import ftw.labels.tests.builders
import logging
import sys


handler = logging.StreamHandler(stream=sys.stderr)
logging.root.addHandler(handler)


class AdaptersZCMLLayer(ComponentRegistryLayer):
    """A layer which only loads the adapters.zcml.
    """

    def setUp(self):
        super(AdaptersZCMLLayer, self).setUp()
        import ftw.labels
        self.load_zcml_file('adapters.zcml', ftw.labels)


ADAPTERS_ZCML_LAYER = AdaptersZCMLLayer()


class LabelsLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        import ftw.labels
        xmlconfig.file('configure.zcml',
                       ftw.labels,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.labels:default')


LABELS_FIXTURE = LabelsLayer()
LABELS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(LABELS_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.labels:functional")



from Products.CMFCore.utils import getToolByName
from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.interfaces import ILabeling
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from ftw.testbrowser import browsing
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from unittest2 import TestCase
from zExceptions import Unauthorized


class TestLabelingView(TestCase):
    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Contributor'])

    @browsing
    def test_activate_labels(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red'),
                                   ('Feature', 'blue')))
        page = create(Builder('labelled page').within(root))
        self.assertFalse(self.indexed_labels_for(page))

        browser.login().open(page,
                             view='labeling/update',
                             data={'activate_labels': ['question', 'bug']})

        self.assertItemsEqual(
            [{'label_id': 'question',
              'title': 'Question',
              'color': 'purple'},
             {'label_id': 'bug',
              'title': 'Bug',
              'color': 'red'}],
            ILabeling(page).active_labels())

        self.assertItemsEqual(['question', 'bug'], self.indexed_labels_for(page))

    @browsing
    def test_deactivate_labels(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red'),
                                   ('Feature', 'blue')))
        page = create(Builder('labelled page')
                      .within(root)
                      .with_labels('question', 'bug'))

        browser.login().open(page,
                             view='labeling/update',
                             data={})

        self.assertItemsEqual(
            [],
            ILabeling(page).active_labels())

    @browsing
    def test_mixed_updating_labels(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red'),
                                   ('Feature', 'blue')))
        page = create(Builder('labelled page')
                      .within(root)
                      .with_labels('question', 'bug'))

        browser.login().open(page,
                             view='labeling/update',
                             data={'activate_labels': ['question','feature']})


        self.assertItemsEqual(
            [{'label_id': 'question',
              'title': 'Question',
              'color': 'purple'},
             {'label_id': 'feature',
              'title': 'Feature',
              'color': 'blue'}],
            ILabeling(page).active_labels())

    @browsing
    def test_updating_is_protected(self, browser):
        root = create(Builder('label root'))
        page = create(Builder('labelled page').within(root))
        browser.login(create(Builder('user').with_roles('Reader')))

        with self.assertRaises(Unauthorized):
            browser.open(page,
                         view='labeling/update',
                         data={'question': 'yes',
                               'feature': 'yes'})

    def indexed_labels_for(self, obj):
        catalog = getToolByName(self.portal, 'portal_catalog')
        rid = catalog.getrid('/'.join(obj.getPhysicalPath()))
        return catalog.getIndexDataForRID(rid).get('labels')



from Acquisition import aq_inner
from Acquisition import aq_parent
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.utils import make_sortable
from OFS.interfaces import IApplication
from persistent.mapping import PersistentMapping
from plone.i18n.normalizer.interfaces import IIDNormalizer
from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.component import adapts
from zope.component import getUtility
from zope.interface import implementer
from zope.interface import implements
from zope.interface import Interface


ANNOTATION_KEY = 'ftw.labels:jar'


class LabelJar(object):
    implements(ILabelJar)
    adapts(ILabelRoot)

    def __init__(self, context):
        self.context = context

    def add(self, title, color):
        label_id = self._make_id(title)
        self.storage[label_id] = PersistentMapping(
            dict(
                label_id=label_id,
                title=title,
                color=color
            )
        )
        return label_id

    def remove(self, label_id):
        if label_id not in self.storage:
            return False

        del self.storage[label_id]
        return True

    def update(self, label_id, title, color):
        self.storage[label_id].update(
            dict(
                title=title,
                color=color
            )
        )

    def get(self, label_id):
        return dict(self.storage[label_id])

    def list(self):
        labels = map(dict, self.storage.values())
        return sorted(labels, key=lambda cls: make_sortable(cls['title']))

    @property
    def storage(self):
        if getattr(self, '_storage', None) is None:
            annotation = IAnnotations(self.context)
            if ANNOTATION_KEY not in annotation:
                annotation[ANNOTATION_KEY] = PersistentMapping()
            self._storage = annotation[ANNOTATION_KEY]
        return self._storage

    def _make_id(self, title):
        normalizer = getUtility(IIDNormalizer)
        label_id = base_id = normalizer.normalize(title)

        counter = 0
        while label_id in self.storage:
            counter += 1
            label_id = '{0}-{1}'.format(base_id, counter)

        return label_id


@implementer(ILabelJar)
@adapter(Interface)
def jar_discovery(context):
    """An ILabelJar adapter for non-ILabelRoot objects, walking
    up the acquisition chain for finding the ILabelJar.
    This allows to adapt any object, which is within an ILabelRoot,
    to ILabelJar without the need to find the root.
    """
    return ILabelJar(aq_parent(aq_inner(context)))


@implementer(ILabelJar)
@adapter(IApplication)
def jar_discovery_app_reached(context):
    """Exits the ILabelJar discovery loop when the Zope app is reached.
    See the jar_discovery docstring for more details.
    """
    raise LookupError('Could not find ILabelJar on any parents.'
                      ' No parent seems to provide ILabelRoot.')



import os
from setuptools import setup, find_packages


version = '1.1.1'


tests_require = [
    'asserts',
    'ftw.builder',
    'ftw.testbrowser',
    'ftw.testing',
    'plone.app.testing',
    'plone.mocktestcase',
    'transaction',
    'unittest2',
    'zope.configuration'
    ]


setup(name='ftw.labels',
      version=version,
      description='A Plone addon for labels.',

      long_description=open('README.rst').read() + '\n' + \
          open(os.path.join('docs', 'HISTORY.txt')).read(),

      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

      keywords='ftw labels',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.labels',

      license='GPL2',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw', ],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'Acquisition',
        'Plone',
        'Products.CMFCore',
        'Products.GenericSetup',
        'ZODB3',
        'Zope2',
        'plone.api',
        'plone.app.portlets',
        'plone.i18n',
        'plone.indexer',
        'plone.portlets',
        'setuptools',
        'zExceptions',
        'zope.annotation',
        'zope.component',
        'zope.i18nmessageid',
        'zope.interface',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



from ftw.builder import Builder
from ftw.builder import create
from ftw.labels.config import COLORS
from ftw.labels.interfaces import ILabelJar
from ftw.labels.testing import LABELS_FUNCTIONAL_TESTING
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import statusmessages
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from unittest2 import TestCase
from zExceptions import BadRequest
from zExceptions import Unauthorized


class TestLabelsJar(TestCase):
    layer = LABELS_FUNCTIONAL_TESTING

    def setUp(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Contributor'])

    @browsing
    def test_create_label(self, browser):
        root = create(Builder('label root'))

        browser.login().open(root,
                             view='labels-jar/create',
                             data={'title': 'Question',
                                   'color': 'purple'})

        self.assertEqual(
            [{'label_id': 'question',
              'title': 'Question',
              'color': 'purple'}],
            ILabelJar(root).list())

    @browsing
    def test_create_label_is_protected(self, browser):
        root = create(Builder('label root'))
        browser.login(create(Builder('user').with_roles('Reader')))

        with self.assertRaises(Unauthorized):
            browser.open(root,
                         view='labels-jar/create',
                         data={'title': 'Question',
                               'color': 'purple'})

    @browsing
    def test_create_label_requires_label_arguments(self, browser):
        root = create(Builder('label root'))

        browser.login().open(root, view='labels-jar/create')

        statusmessages.assert_message(u"Please choose a title.")

    @browsing
    def test_create_label_without_color_add_random_existing_color(self, browser):
        root = create(Builder('label root'))
        browser.login()
        for i in range(0, len(COLORS)):
            browser.open(root,
                         view='labels-jar/create',
                         data={'title': 'Question'})

        selected_colors = [
            label.get('color') for label in ILabelJar(root).list()]
        self.assertEquals(len(COLORS), len(list(set(selected_colors))))

    @browsing
    def test_update_label(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red')))

        browser.login().open(root,
                             view='labels-jar/update',
                             data={'label_id': 'question',
                                   'title': 'Questions and inquiries',
                                   'color': 'green'})

        self.assertItemsEqual(
            [{'label_id': 'question',
              'title': 'Questions and inquiries',
              'color': 'green'},

             {'label_id': 'bug',
              'title': 'Bug',
              'color': 'red'}],
            ILabelJar(root).list())

    @browsing
    def test_update_label_requires_label_id(self, browser):
        root = create(Builder('label root'))

        with self.assertRaises(BadRequest) as cm:
            browser.login().open(root, view='labels-jar/update')

        self.assertEquals('The "label_id" request argument is required.',
                          str(cm.exception))

    @browsing
    def test_partial_update_label(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red')))

        browser.login().open(root,
                             view='labels-jar/update',
                             data={'label_id': 'question',
                                   'title': 'Questions and inquiries'})

        self.assertItemsEqual(
            [{'label_id': 'question',
              'title': 'Questions and inquiries',
              'color': 'purple'},

             {'label_id': 'bug',
              'title': 'Bug',
              'color': 'red'}],
            ILabelJar(root).list())

    @browsing
    def test_remove_label(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Question', 'purple'),
                                   ('Bug', 'red')))

        browser.login().open(root,
                             view='labels-jar/remove',
                             data={'label_id': 'question'})

        self.assertEqual(
            [{'label_id': 'bug',
              'title': 'Bug',
              'color': 'red'}],
            ILabelJar(root).list())

    @browsing
    def test_remove_label_requires_label_id(self, browser):
        root = create(Builder('label root'))

        with self.assertRaises(BadRequest) as cm:
            browser.login().open(root, view='labels-jar/remove')

        self.assertEquals('The "label_id" request argument is required.',
                          str(cm.exception))

    @browsing
    def test_edit_label_form_change_title(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Feature', 'blue')))

        browser.login().open(root,
                             view='labels-jar/edit_label',
                             data={'label_id': 'feature'})

        browser.fill({'title': 'Features and inquiries'}).submit()

        self.assertEqual(
            [{'label_id': 'feature',
              'title': 'Features and inquiries',
              'color': 'blue'}],
            ILabelJar(root).list())

    @browsing
    def test_edit_label_form_change_color(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Feature', 'blue')))

        browser.login().open(root,
                             view='labels-jar/edit_label',
                             data={'label_id': 'feature'})

        browser.fill({'color': 'green'}).submit()

        self.assertEqual(
            [{'label_id': 'feature',
              'title': 'Feature',
              'color': 'green'}],
            ILabelJar(root).list())

    @browsing
    def test_edit_label_form_delete_label(self, browser):
        root = create(Builder('label root')
                      .with_labels(('Feature', 'blue')))

        browser.login().open(root,
                             view='labels-jar/edit_label',
                             data={'label_id': 'feature'})

        browser.find('Delete label').click()

        self.assertEqual([], ILabelJar(root).list())



from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from plone.app.layout.viewlets.common import ViewletBase
from ftw.labels.interfaces import ILabeling
from ftw.labels.interfaces import ILabelSupport
from Products.CMFCore.utils import getToolByName


class LabelingViewlet(ViewletBase):

    index = ViewPageTemplateFile('labeling.pt')

    @property
    def available(self):
        if not self.available_labels:
            return False
        if 'portal_factory' in self.context.absolute_url():
            return False
        if not ILabelSupport.providedBy(self.context):
            return False
        if not tuple(self.available_labels):
            return False
        return True

    @property
    def active_labels(self):
        return ILabeling(self.context).active_labels()

    @property
    def available_labels(self):
        return ILabeling(self.context).available_labels()

    @property
    def can_edit(self):
        mtool = getToolByName(self.context, 'portal_membership')
        return mtool.checkPermission('ftw.labels: Change Labels', self.context)



from ftw.labels import _
from ftw.labels.config import COLORS
from ftw.labels.interfaces import ILabelJar
from plone import api
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zExceptions import BadRequest
import random


class LabelsJar(BrowserView):

    edit_label_template = ViewPageTemplateFile('templates/edit_label.pt')

    def create(self):
        """Create a new label.
        """

        title = self.request.form.get('title', None)
        if not title:
            api.portal.show_message(
                _(u'lable_title_is_missing',
                  default=u'Please choose a title.'),
                self.request, 'error')

            return self._redirect()

        color = self.request.form.get('color', None)
        if not color:
            color = self._get_random_color()

        jar = ILabelJar(self.context)
        jar.get(jar.add(title, color))
        return self._redirect()

    def update(self):
        """Update a label.
        """

        if self.request.form.get('form.delete', None):
            return self.remove()

        label_id = self.request.form.get('label_id', None)
        if not label_id:
            raise BadRequest('The "label_id" request argument is required.')

        jar = ILabelJar(self.context)
        label = jar.get(label_id)

        title = self.request.form.get('title', None)
        if title:
            label['title'] = title

        color = self.request.form.get('color', None)
        if color:
            label['color'] = color

        jar.update(**label)
        return self._redirect()

    def remove(self):
        """Remove a label.
        """

        label_id = self.request.form.get('label_id', None)
        if not label_id:
            raise BadRequest('The "label_id" request argument is required.')

        ILabelJar(self.context).remove(label_id)
        return self._redirect()

    def edit_label(self):
        """Form for editing a label.
        """
        return self.edit_label_template()

    def get_label(self):
        label_id = self.request.form.get('label_id', None)
        if not label_id:
            raise BadRequest('The "label_id" request argument is required.')
        return ILabelJar(self.context).get(label_id)

    @property
    def colors(self):
        return [dict(
            normal=color,
            light='{0}-light'.format(color)
            ) for color in COLORS]

    def _redirect(self):
        response = self.request.RESPONSE
        return response.redirect(self.context.absolute_url())

    def _get_random_color(self):
        all_colors = list(COLORS)
        all_colors.extend(
            ['{0}-light'.format(color) for color in COLORS])

        used_colors = [
            label.get('color') for label in ILabelJar(self.context).list()]

        available_colors = tuple(set(all_colors) - set(used_colors))

        if not available_colors:
            available_colors = all_colors

        return random.choice(available_colors)



from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.jar import LabelJar
from ftw.labels.testing import ADAPTERS_ZCML_LAYER
from zope.annotation import IAttributeAnnotatable
from zope.component import queryAdapter
from plone.mocktestcase.dummy import Dummy
from zope.interface.verify import verifyClass
from unittest2 import TestCase
from zope.interface import alsoProvides


class TestLabelJar(TestCase):
    layer = ADAPTERS_ZCML_LAYER

    def setUp(self):
        self.root = Dummy()
        alsoProvides(self.root, ILabelRoot, IAttributeAnnotatable)

    def test_label_jar_implements_interface(self):
        self.assertTrue(ILabelJar.implementedBy(LabelJar),
                        'LabelJar should implement ILabelJar')
        verifyClass(ILabelJar, LabelJar)

    def test_adapter_is_registered_for_label_root(self):
        self.assertTrue(
            queryAdapter(self.root, ILabelJar),
            'The LabelJar adapter is not registered for ILabelRoot')

    def test_adding_new_label(self):
        jar = LabelJar(self.root)
        label_id = jar.add('Question', '#FF0000')

        self.assertDictEqual({'label_id': label_id,
                              'title': 'Question',
                              'color': '#FF0000'},
                             jar.get(label_id))

    def test_label_ids_are_unique(self):
        jar = LabelJar(self.root)

        first_label_id = jar.add('Question', '#FF0000')
        second_label_id = jar.add('Question', '#FF0000')

        self.assertNotEquals(second_label_id, first_label_id,
                             'Labels ID should be unique.')

    def test_listing_labels(self):
        jar = LabelJar(self.root)

        first_label_id = jar.add('First', '#FF0000')
        second_label_id = jar.add('Second', '#0000FF')

        self.assertItemsEqual(
            [{'label_id': first_label_id,
              'title': 'First',
              'color': '#FF0000'},
             {'label_id': second_label_id,
              'title': 'Second',
              'color': '#0000FF'}],
            jar.list())

    def test_listing_labels_is_sorted(self):
        jar = LabelJar(self.root)

        jar.add('Zeta-0', '')
        jar.add('zeta-1', '')
        jar.add('alpha-0', '')
        jar.add('\xc3\x84lpha-1', '')
        jar.add('Alpha-2', '')

        self.assertEqual(
            ['alpha-0', '\xc3\x84lpha-1', 'Alpha-2', 'Zeta-0', 'zeta-1'],
            [label.get('title') for label in jar.list()])

    def test_updating_labels(self):
        jar = LabelJar(self.root)

        label_id = jar.add('Question', '#FF0000')
        self.assertDictEqual({'label_id': label_id,
                              'title': 'Question',
                              'color': '#FF0000'},
                             jar.get(label_id))

        jar.update(label_id, 'New Question', '#0000FF')
        self.assertDictEqual({'label_id': label_id,
                              'title': 'New Question',
                              'color': '#0000FF'},
                             jar.get(label_id))

    def test_remove_labels(self):
        jar = LabelJar(self.root)

        label_id = jar.add('Question', '#FF0000')
        self.assertEqual(
            [{'label_id': label_id,
              'title': 'Question',
              'color': '#FF0000'}],
            jar.list())

        jar.remove(label_id)
        self.assertEqual([], jar.list())

    def test_label_dict_mutations_are_not_stored(self):
        jar = LabelJar(self.root)
        label_id = jar.add('Question', '#FF0000')

        label = jar.get(label_id)
        label['title'] = 'HACK THE TITLE'

        self.assertDictEqual({'label_id': label_id,
                              'title': 'Question',
                              'color': '#FF0000'},
                             jar.get(label_id))

    def test_list_mutations_are_not_stored(self):
        jar = LabelJar(self.root)
        label_id = jar.add('Question', '#FF0000')

        labels = jar.list()
        self.assertEqual(
            [{'label_id': label_id,
              'title': 'Question',
              'color': '#FF0000'}],
            labels)

        del labels[0]
        self.assertEqual(
            [{'label_id': label_id,
              'title': 'Question',
              'color': '#FF0000'}],
            jar.list())

    def test_data_is_stored_persistently(self):
        jar = LabelJar(self.root)

        label_id = jar.add('Question', '#FF0000')
        self.assertEqual(
            [{'label_id': label_id,
              'title': 'Question',
              'color': '#FF0000'}],
            jar.list())

        jar = LabelJar(self.root)
        self.assertEqual(
            [{'label_id': label_id,
              'title': 'Question',
              'color': '#FF0000'}],
            jar.list())



from zope.interface import Interface


class ILabelRoot(Interface):
    """Marker interface for label roots, which have their own ecosystem
    of labels.
    """


class ILabelSupport(Interface):
    """Marker interface for objects which support labelling.
    All objects providing this interface must be within a container
    providing ``ILabelRoot``.
    """


class ILabelJarChild(Interface):
    """Marker interface for objects which should display the labelsjar
    portlet of a parents ``ILabelRoot``
    """


class ILabelJar(Interface):
    """Adapter interface for label roots (``ILabelRoot``), providing
    an interface for managing available labels for this label ecosystem.
    """

    def __init__(context):
        """Adapts a label root.
        """

    def add(title, color):
        """Adds a new label with ``title`` and ``color``.
        Returns the id of the new label as string.
        """

    def remove(label_id):
        """Removes the label with ID ``label_id``.
        If the label was successfully removed, ``True`` is returned.
        If there is no such label, ``False`` is returned.
        """

    def update(label_id, title, color):
        """Updates the label with ID ``label_id``.
        Raises a ``KeyError`` when there is no such label.
        """

    def get(label_id):
        """Returns a dict with the label information for the label
        with ID ``label_id``.

        Example:

        {'label_id': 'label-id',
         'title': 'Label Title',
         'color': '#FF0000'}
        """

    def list():
        """Returns all available labels as list of dicts.

        Example:
        [
            {'label_id': 'label-id',
             'title': 'Label Title',
             'color': '#FF0000'},

            {'label_id': 'another-label-id',
             'title': 'Another Label Title',
             'color': '#0000FF'},
        ]
        """


class ILabeling(Interface):
    """Adapter interface for objects which provide ``ILabelSupport``,
    providing an interface for setting, updating and removing the
    labels of this object.
    """

    def __init__(context):
        """Adapts any object with ``ILabelSupport``.
        """

    def update(label_ids):
        """Set the active labels for the current context by passing in a list
        of label IDs.
        Labels which where activated before but are not in the ``label_ids``
        are purged.
        """

    def active_labels():
        """Returns all active labels on the current object as list of dicts.

        Example:
        [
            {'label_id': 'label-id',
             'title': 'Label Title',
             'color': '#FF0000'},

            {'label_id': 'another-label-id',
             'title': 'Another Label Title',
             'color': '#0000FF'},
        ]
        """

    def available_labels():
        """Returns all labels available in the current ecosystem, each as
        dict, with additional information whether it is activated on the
        current object or not.

        Example:
        [
            {'label_id': 'label-id',
             'title': 'Label Title',
             'color': '#FF0000',
             'active': True},

            {'label_id': 'another-label-id',
             'title': 'Another Label Title',
             'color': '#0000FF',
             'active': False},
        ]
        """
